#!/usr/bin/env python

# (C) 2002-2009 Chris Liechti <cliechti@gmx.net>
# redirect data from a TCP/IP connection to a serial port and vice versa
# requires Python 2.2 'cause socket.sendall is used

# (C) 2010 Ben Kempke <bpkempke@umich.edu>
# modified to dump all serial activity to file


import sys
import os
import time
import threading
import socket
import codecs
import time
import datetime
import serial
import base64

class Redirector:
    def __init__(self, serial_instance, socket, srv, stationname, radioid, ser_newline=None, net_newline=None, spy=False):
        self.serial = serial_instance
        self.socket = socket
        self.ser_newline = ser_newline
        self.net_newline = net_newline
        self.spy = spy
        self.srv = srv
        self.stationname = stationname
        self.lastsavedtime = 0
        self._write_lock = threading.Lock()
        self._historic_data_lock = threading.Lock()
        self.historic_data = str()
        self.radio_id = radioid

    def shortcut(self):
        """connect the serial port to the TCP port by copying everything
           from one side to the other"""
        self.alive = True
        self.thread_read = threading.Thread(target=self.reader)
        self.thread_read.setDaemon(True)
        self.thread_read.setName('serial->socket')
        self.thread_read.start()
		
		#Set up a logging thread which checks the 
        self.thread_log = threading.Thread(target=self.logger)
        self.thread_log.setDaemon(True)
        self.thread_log.setName('serial->logger')
        self.thread_log.start()
        
        self.writer()

    def setconnection(self, socket):
        self.socket = socket
		
    def logger(self):
        while self.alive:
            time.sleep(1)
            
            # Need to make sure two threads aren't modifying historic_data at the same time
            self._historic_data_lock.acquire()
            if self.historic_data:
                #Log the incoming data, as a hex string
                today = datetime.date.today()
                todaystring = today.strftime("%m%d%y")
                fid = open(self.stationname + "_serial_log_" + todaystring + ".dat", "a")
                utct = datetime.datetime.utcnow()
                t = time.mktime(utct.timetuple())      # append the timestamp of this incoming data
                self.lastsavedtime = t
                fid.write("\n\n" + utct.strftime("%m/%d/%Y %I:%M:%S (") + str(self.radio_id) + ", " + str(len(self.historic_data)) + ")\n")
                fid.write(base64.b64encode(self.historic_data))
                #fid.write(''.join( [ "%02X" % ord( x ) for x in data ] ).strip())
                fid.close()
                self.historic_data = str();
            self._historic_data_lock.release()

    def reader(self):
        """loop forever and copy serial->socket"""
        while self.alive:
            try:
                data = self.serial.read(1)              # read one, blocking
                n = self.serial.inWaiting()             # look if there is more
                if n:
                    data = data + self.serial.read(n)   # and get as much as possible
                if data:
					
					# append the currently-read data onto the historic data array
                    self._historic_data_lock.acquire()
                    self.historic_data += data
                    self._historic_data_lock.release()
					
                    # the spy shows what's on the serial port, so log it before converting newlines
                    if self.spy:
                        sys.stdout.write(codecs.escape_encode(data)[0])
                        sys.stdout.flush()
                    if self.ser_newline and self.net_newline:
                        # do the newline conversion
                        # XXX fails for CR+LF in input when it is cut in half at the begin or end of the string
                        data = net_newline.join(data.split(ser_newline))
                    
                    if self.socket:
                        # escape outgoing data when needed (Telnet IAC (0xff) character)
                        self._write_lock.acquire()
                        try:
                            self.socket.sendall(data)           # send it over TCP
                        finally:
                            self._write_lock.release()
            except socket.error:
                sys.stderr.write('ERROR: socket disconnected')
                # probably got disconnected
                break
        self.alive = False

#    def write(self, data):
#        """thread safe socket write with no data escaping. used to send telnet stuff"""
 #       self._write_lock.acquire()
 #       try:
 #           self.socket.sendall(data)
 #       finally:
 #           self._write_lock.release()

    def writer(self):
        """loop forever and copy socket->serial"""
        while self.alive:
            if self.socket:
                try:
                    data = self.socket.recv(1024)
                    if not data:
                        break
                    if self.ser_newline and self.net_newline:
                        # do the newline conversion
                        # XXX fails for CR+LF in input when it is cut in half at the begin or end of the string
                        data = ser_newline.join(data.split(net_newline))
                    self.serial.write(data)                 # get a bunch of bytes and send them
                    # the spy shows what's on the serial port, so log it after converting newlines
                    if self.spy:
                        sys.stdout.write(codecs.escape_encode(data)[0])
                        sys.stdout.flush()
                except socket.error:
                    sys.stderr.write('ERROR: socket disconnected\n')
                    # probably got disconnected
                    break
            else:
                self.socket, addr = self.srv.accept()
                sys.stderr.write('Connected by %s\n' % (addr,))
        self.alive = False
        self.thread_read.join()

    def stop(self):
        sys.stdout.write("stopping")
        """Stop copying"""
        if self.alive:
            self.alive = False
            self.thread_read.join()


if __name__ == '__main__':
    import optparse

    parser = optparse.OptionParser(
        usage = "%prog [options] [port [baudrate]]",
        description = "Simple Serial to Network (TCP/IP) redirector.",
        epilog = """\
NOTE: no security measures are implemented. Anyone can remotely connect
to this service over the network.

Only one connection at once is supported. When the connection is terminated
it waits for the next connect.
""")

    parser.add_option("-q", "--quiet",
        dest = "quiet",
        action = "store_true",
        help = "suppress non error messages",
        default = False
    )

    parser.add_option("--spy",
        dest = "spy",
        action = "store_true",
        help = "peek at the communication and print all data to the console",
        default = False
    )

    group = optparse.OptionGroup(parser,
        "Serial Port",
        "Serial port settings"
    )
    parser.add_option_group(group)

    group.add_option("-p", "--port",
        dest = "port",
        help = "port, a number (default 0) or a device name",
        default = None
    )

    group.add_option("-b", "--baud",
        dest = "baudrate",
        action = "store",
        type = 'int',
        help = "set baud rate, default: %default",
        default = 9600
    )

    group.add_option("", "--parity",
        dest = "parity",
        action = "store",
        help = "set parity, one of [N, E, O], default=%default",
        default = 'N'
    )

    group.add_option("--rtscts",
        dest = "rtscts",
        action = "store_true",
        help = "enable RTS/CTS flow control (default off)",
        default = False
    )

    group.add_option("--xonxoff",
        dest = "xonxoff",
        action = "store_true",
        help = "enable software flow control (default off)",
        default = False
    )

    group.add_option("--rts",
        dest = "rts_state",
        action = "store",
        type = 'int',
        help = "set initial RTS line state (possible values: 0, 1)",
        default = None
    )

    group.add_option("--dtr",
        dest = "dtr_state",
        action = "store",
        type = 'int',
        help = "set initial DTR line state (possible values: 0, 1)",
        default = None
    )

    group = optparse.OptionGroup(parser,
        "Network settings",
        "Network configuration."
    )
    parser.add_option_group(group)

    group.add_option("-P", "--localport",
        dest = "local_port",
        action = "store",
        type = 'int',
        help = "local TCP port",
        default = 7777
    )

    group = optparse.OptionGroup(parser,
        "Newline Settings",
        "Convert newlines between network and serial port. Conversion is normally disabled and can be enabled by --convert."
    )
    parser.add_option_group(group)

    group.add_option("-c", "--convert",
        dest = "convert",
        action = "store_true",
        help = "enable newline conversion (default off)",
        default = False
    )
	
    group.add_option("-n", "--name",
        dest = "stationname",
        action = "store",
        help = "set station name",
        default = "default"
    )
    
    group.add_option("-i", "--id",
        dest = "radioid",
        action = "store",
        help = "set radio id",
        default = 0
    )

    group.add_option("--net-nl",
        dest = "net_newline",
        action = "store",
        help = "type of newlines that are expected on the network (default: %default)",
        default = "LF"
    )

    group.add_option("--ser-nl",
        dest = "ser_newline",
        action = "store",
        help = "type of newlines that are expected on the serial port (default: %default)",
        default = "CR+LF"
    )

    (options, args) = parser.parse_args()

    # get port and baud rate from command line arguments or the option switches
    port = options.port
    baudrate = options.baudrate
    if args:
        if options.port is not None:
            parser.error("no arguments are allowed, options only when --port is given")
        port = args.pop(0)
        if args:
            try:
                baudrate = int(args[0])
            except ValueError:
                parser.error("baud rate must be a number, not %r" % args[0])
            args.pop(0)
        if args:
            parser.error("too many arguments")
    else:
        if port is None: port = 0

    # check newline modes for network connection
    mode = options.net_newline.upper()
    if mode == 'CR':
        net_newline = '\r'
    elif mode == 'LF':
        net_newline = '\n'
    elif mode == 'CR+LF' or mode == 'CRLF':
        net_newline = '\r\n'
    else:
        parser.error("Invalid value for --net-nl. Valid are 'CR', 'LF' and 'CR+LF'/'CRLF'.")

    # check newline modes for serial connection
    mode = options.ser_newline.upper()
    if mode == 'CR':
        ser_newline = '\r'
    elif mode == 'LF':
        ser_newline = '\n'
    elif mode == 'CR+LF' or mode == 'CRLF':
        ser_newline = '\r\n'
    else:
        parser.error("Invalid value for --ser-nl. Valid are 'CR', 'LF' and 'CR+LF'/'CRLF'.")

    # connect to serial port
    ser = serial.Serial()
    ser.port     = port
    ser.baudrate = baudrate
    ser.parity   = options.parity
    ser.rtscts   = options.rtscts
    ser.xonxoff  = options.xonxoff
    ser.timeout  = 1     # required so that the reader thread can exit

    if not options.quiet:
        sys.stderr.write("--- TCP/IP to Serial redirector --- type Ctrl-C / BREAK to quit\n")
        sys.stderr.write("--- %s %s,%s,%s,%s ---\n" % (ser.portstr, ser.baudrate, 8, ser.parity, 1))

    try:
        ser.open()
    except serial.SerialException:
        sys.stderr.write("Could not open serial port %s: \n" % (ser.portstr))
        sys.exit(1)

    if options.rts_state is not None:
        ser.setRTS(options.rts_state)

    if options.dtr_state is not None:
        ser.setDTR(options.dtr_state)

    srv = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    srv.bind( ('', options.local_port) )
    srv.listen(1)
    while True:
        try:
            sys.stderr.write("Waiting for connection on %s...\n" % options.local_port)
            connection = None
            
            # enter network <-> serial loop
            r = Redirector(
                ser,
                connection,
				srv,
                options.stationname,
                options.radioid,
                options.convert and ser_newline or None,
                options.convert and net_newline or None,
                options.spy,
            )
            r.shortcut()
#            connection, addr = srv.accept()
#            r.setconnection(connection)
#            sys.stderr.write('Connected by %s\n' % (addr,))
            if options.spy: sys.stdout.write('\n')
            sys.stderr.write('Disconnected\n')
#            connection.close()
        except KeyboardInterrupt:
            break
        except socket.error:
            sys.stderr.write('ERROR: socket disconnect\n')

    sys.stderr.write('\n--- exit ---\n')

